'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var command = require('@oclif/command');
var cli = _interopDefault(require('cli-ux'));
var npm = _interopDefault(require('npm'));
var once = _interopDefault(require('lodash/once'));
var util = require('util');
var fs = require('fs');
var path = _interopDefault(require('path'));
var readPackageTree = _interopDefault(require('read-package-tree'));
var semver = _interopDefault(require('semver'));
var npm$1 = _interopDefault(require('npm/lib/npm.js'));
var mutateIntoLogicalTree = _interopDefault(require('npm/lib/install/mutate-into-logical-tree.js'));
var deps_js = require('npm/lib/install/deps.js');
var readShrinkwrap = _interopDefault(require('npm/lib/install/read-shrinkwrap.js'));
var packageId = _interopDefault(require('npm/lib/utils/package-id.js'));

const states = {
  scanTree: 0,
  readLockfile: 1,
  verify: 2,
};

function noop() {}

async function check(dir, options = {development: false, production: false, ignorePeer: false}) {
  const progress = options.progress || noop;
  return new Promise((resolve, reject) => {
    progress(states.scanTree);
    readPackageTree(dir, function(err, physicalTree) {
      if (err) {
        reject(err);
        return;
      }

      if (!physicalTree) physicalTree = {package: {}, path: dir};
      physicalTree.isTop = true;
      progress(states.readLockfile);
      // Inflation is necessary to catch errors like extraneous packages and some
      // kinds of missing dependencies.
      readShrinkwrap.andInflate(physicalTree, function(err) {
        if (!err) {
          try {
            progress(states.verify);
            resolve(checkDependencies(dir, deps_js.computeMetadata(physicalTree), options));
            return;
          } catch (e) {
            err = e;
          }
        }
        reject(err);
      });
    });
  });
}

function inList(list, value) {
  return list.indexOf(value) !== -1;
}

function checkDependencies(dir, physicalTree, options) {
  const data = mutateIntoLogicalTree.asReadInstalled(physicalTree);

  pruneNestedExtraneous(data);
  filterByEnv(data, options);

  const unlooped = filterFound(unloop(data));
  const lite = getLite(unlooped, undefined, {ignorePeer: options.ignorePeer});

  // const allDependencies = new Map();

  // gatherDependencies(unlooped);
  // console.log(unlooped.dependencies['synchronize-pool']);

  return lite.problems || [];
}

// function gatherDependencies(data) {
//   if (hasOwnProperty.call(data, '_found') && data._found !== 'explicit')
// }

function pruneNestedExtraneous(data, visited) {
  visited = visited || [];
  visited.push(data);
  for (const i in data.dependencies) {
    if (data.dependencies[i].extraneous) {
      data.dependencies[i].dependencies = {};
    } else if (visited.indexOf(data.dependencies[i]) === -1) {
      pruneNestedExtraneous(data.dependencies[i], visited);
    }
  }
}

function filterByEnv(data, {development, production}) {
  const dependencies = {};
  const devKeys = Object.keys(data.devDependencies || []);
  const prodKeys = Object.keys(data._dependencies || []);
  Object.keys(data.dependencies).forEach(function(name) {
    if (!development && inList(devKeys, name) && !inList(prodKeys, name) && data.dependencies[name].missing) {
      return;
    }

    if ((development && inList(devKeys, name)) || // only --dev
        (production && inList(prodKeys, name)) || // only --production
        (!development && !production)) { // no --production|--dev|--only=xxx
      dependencies[name] = data.dependencies[name];
    }
  });
  data.dependencies = dependencies;
}

function isCruft(data) {
  return data.extraneous && data.error && data.error.code === 'ENOTDIR';
}

function getLite(data, noname, {ignorePeer = false} = {}) {
  const lite = {};

  if (isCruft(data)) return lite;

  if (!noname && data.name) lite.name = data.name;
  if (data.version) lite.version = data.version;
  if (data.extraneous) {
    lite.extraneous = true;
    lite.problems = lite.problems || [];
    lite.problems.push('extraneous: ' + packageId(data) + ' ' + (data.path || ''));
  }

  if (data.error && data.path !== path.resolve(npm$1.globalDir, '..') &&
      (data.error.code !== 'ENOENT' || noname)) {
    lite.invalid = true;
    lite.problems = lite.problems || [];
    const message = data.error.message;
    lite.problems.push('error in ' + data.path + ': ' + message);
  }

  if (data._from) {
    lite.from = data._from;
  }

  if (data._resolved) {
    lite.resolved = data._resolved;
  }

  if (data.invalid) {
    lite.invalid = true;
    lite.problems = lite.problems || [];
    lite.problems.push('invalid: ' +
                       packageId(data) +
                       ' ' + (data.path || ''));
  }

  if (data.peerInvalid) {
    lite.peerInvalid = true;
    if (!ignorePeer) {
      lite.problems = lite.problems || [];
      lite.problems.push('peer dep not met: ' +
                         packageId(data) +
                         ' ' + (data.path || ''));
    }
  }

  const deps = (data.dependencies && Object.keys(data.dependencies)) || [];
  if (deps.length) {
    lite.dependencies = deps.map(function(d) {
      const dep = data.dependencies[d];
      if (dep.missing && !dep.optional) {
        lite.problems = lite.problems || [];
        let p = 'missing: ';
        p += d + '@' + dep.requiredBy +
            ', required by ' +
            packageId(data);
        lite.problems.push(p);
        if (dep.dependencies) {
          return [d, getLite(dep, true, {ignorePeer})];
        } else {
          return [d, { required: dep.requiredBy, missing: true }];
        }
      } else if (dep.peerMissing) {
        if (!ignorePeer) {
          lite.problems = lite.problems || [];
          dep.peerMissing.forEach(function(missing) {
            const pdm = 'peer dep missing: ' +
                missing.requires +
                ', required by ' +
                missing.requiredBy;
            lite.problems.push(pdm);
          });
        }
        return [d, { required: dep, peerMissing: true }];
      }
      return [d, getLite(dep, true, {ignorePeer})];
    }).reduce(function(deps, d) {
      if (d[1].problems) {
        lite.problems = lite.problems || [];
        lite.problems.push.apply(lite.problems, d[1].problems);
      }
      deps[d[0]] = d[1];
      return deps;
    }, {});
  }
  return lite;
}

function unloop(root) {
  const queue = [root];
  const seen = new Set();
  seen.add(root);

  while (queue.length) {
    const current = queue.shift();
    const deps = current.dependencies = current.dependencies || {};
    Object.keys(deps).forEach(function(d) {
      let dep = deps[d];
      if (dep.missing && !dep.dependencies) return;
      if (dep.path && seen.has(dep)) {
        dep = deps[d] = Object.assign({}, dep);
        dep.dependencies = {};
        dep._deduped = path.relative(root.path, dep.path).replace(/node_modules\//g, '');
        return;
      }
      seen.add(dep);
      queue.push(dep);
    });
  }

  return root;
}

function filterFound(root) {
  const args = [];
  if (!args.length) return root;
  if (!root.dependencies) return root;

  // Mark all deps
  const toMark = [root];
  while (toMark.length) {
    const markPkg = toMark.shift();
    const markDeps = markPkg.dependencies;
    if (!markDeps) continue;
    Object.keys(markDeps).forEach(function(depName) {
      const dep = markDeps[depName];
      if (dep.peerMissing) return;
      dep._parent = markPkg;
      for (let ii = 0; ii < args.length; ii++) {
        const argName = args[ii][0];
        const argVersion = args[ii][1];
        const argRaw = args[ii][2];
        let found;
        if (typeof argRaw === 'object') {
          if (dep.path === argRaw.path) {
            found = true;
          }
        } else if (depName === argName && argVersion) {
          found = semver.satisfies(dep.version, argVersion, true);
        } else if (depName === argName) {
          // If version is missing from arg, just do a name match.
          found = true;
        }
        if (found) {
          dep._found = 'explicit';
          let parent = dep._parent;
          while (parent && !parent._found && !parent._deduped) {
            parent._found = 'implicit';
            parent = parent._parent;
          }
          break;
        }
      }
      toMark.push(dep);
    });
  }
  const toTrim = [root];
  while (toTrim.length) {
    const trimPkg = toTrim.shift();
    const trimDeps = trimPkg.dependencies;
    if (!trimDeps) continue;
    trimPkg.dependencies = {};
    Object.keys(trimDeps).forEach(function(name) {
      const dep = trimDeps[name];
      if (!dep._found) return;
      if (dep._found === 'implicit' && dep._deduped) return;
      trimPkg.dependencies[name] = dep;
      toTrim.push(dep);
    });
  }
  return root;
}

const statAsync = util.promisify(fs.stat);

const loadNPM = once(() => npm.load());

const stateMessages = new Map([
  [states.scanTree, 'scanning tree'],
  [states.readLockfile, 'reading lockfile'],
  [states.verify, 'verifying consistency'],
]);

class CheckDependenciesExceptPeerCommand extends command.Command {
  async run() {
    const {
      args: {
        directory: dir = process.cwd(),
      },
      flags
    } = this.parse(CheckDependenciesExceptPeerCommand);
    cli.action.start('initializing');
    if (!(await statAsync(dir)).isDirectory()) {
      this.error('the provided path is not a directory', {exit: 1});
    }
    loadNPM();
    const errors = await check(dir, {
      ignorePeer: flags['ignore-peer'],
      progress(state) {
        const message = stateMessages.get(state) || 'looking around for breadcrumbs';
        cli.action.start(message);
      },
    });
    if (errors && errors.length) {
      this.error(errors.join('\n'), {exit: 2});
    }
  }
}

CheckDependenciesExceptPeerCommand.description = `Verify the lockfile and installed dependencies, and ignore unmatched peer dependencies
Scans the lockfile and installed dependencies for inconsistencies. Ignores peer dependency violations by default.
`;

CheckDependenciesExceptPeerCommand.args = [{
  name: 'directory',
  description: '[default: PWD] path to the root of the module to check, defaults to the working directory',
  required: false,
}];

CheckDependenciesExceptPeerCommand.flags = {
  // add --version flag to show CLI version
  version: command.flags.version({char: 'v'}),
  // add --help flag to show CLI version
  help: command.flags.help({char: 'h'}),
  'ignore-peer': command.flags.boolean({
    char: 'i',
    description: 'whether to ignore peer dependencies',
    default: true,
  }),
  dev: command.flags.boolean({
    char: 'D',
    description: 'whether to include development dependencies (default both)',
  }),
  prod: command.flags.boolean({
    char: 'P',
    description: 'whether to include production dependencies (default both)',
  }),
};

module.exports = CheckDependenciesExceptPeerCommand;
